// --- [parser.c] -------------------------------------------------------------
//
// tab = 3
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "inttypes.h"
#include <windows.h>

#define ESCAPE       92  // '\' will nicht
#define SEPARATOR    ','

#define FORCE_1_BYTE        'S'     // nachgestelltes Zeichen erzwingt Datenlnge
#define FORCE_2_BYTE        'L'     // kein Hexadezimalzeichen verwenden !!
#define FORCE_4_BYTE        'W'

enum { UNKNOWN, UNSIGNED_DEZ, POSITIV_DEZ, NEGATIV_DEZ, HEXADEZIMAL, OKTAL, BINAER, STRING};

unsigned int test_numstring(char* s);
unsigned int bin_2_byte(char * cv);
unsigned int oct_2_byte(char * cv);
unsigned int dez_2_byte(char * cv);
unsigned int hex_2_byte(char * cv);

// ----------------------------------------------------------------------------
// Nimmt eine Zeichenfolge 'string' entgegen, zerlegt sie in die Teile, die durch Kommata eingeschlossen sind
// und versucht diese Bereiche als Zahl zu interpretieren
// es wird gepft, ob
// - alle Zeichen Ziffern von 0..9 sind -> Dezimalzahl
// - das erste Zeichen ein '-' und der Rest Ziffern von 0..9 sind -> negative Dezimalzahl
// - das erste Zeichen eine '0' (Null) ist und
//   - das zweite Zeichen ein 'x/X' und der Ziffern von 0..F sind -> Hexadezimalzahl
//   - das zweite Zeichen ein 'b/B' und der Ziffern von 0..1 sind -> Binaerzahl
//  das erste Zeichen ein 'o/O' und der Rest Ziffern von 0..7 sind -> Oktalzahl
// Fallen alle Tests negativ aus, wird der String als Textstring bernommen.
//
// Das Ergebnis der Umwandlung wird in result[] abgelegt, die Lnge dieses Strings
// (der nicht durch \0 terminiert sein kann) steht in str_len
//
// Wenn Zahlen erkannt wurden, die nicht in eine 8Bit-Variable passen, dann werden
// automatisch 2 bzw. 4 Byte fr das Ergebnis zurckgegeben.
// Das Anhngen eines 'l/L' an eine Ziffernfolge (gleich welchen Formats) erzwingt die
// Rckgabe eines 16-Bit-Wertes, auch wenn 8-Bit ausreichend wren.
// Das Anhngen eines 'w/W' erzwingt die Rckgabe eines 32-Bit Wertes,
// auch wenn 16- bzw.8-Bit ausreichend wren.
//
// unsigned   : 123..890  -> als reine Folge aus Ziffern
// integer    : -2        -> wie unsigned, aber mit fhrendem '-' 0xFE
// hexadezimal: 0x12BC    -> fhrendes "0x" und Zeichenfolge "0123...CDEF"
// ein nachfolgendes 's/S' erzeugt eine  8Bit Zahl bei allen Zahlenformaten.
// ein nachfolgendes 'l/L' erzeugt eine 16Bit Zahl bei allen Zahlenformaten:
// 128l/L     : -> 0080   dagegen 128   : -> 80
// -128l      : -> 0xFF80 dagegen -128  : ->80
// 0x0Fl      : -> 000F   dagegen 0F    : -> 0F
// ein nachfolgendes 'w/W' erzeugt eine 32Bit Zahl bei allen Zahlenformaten:
// 128w/W     : -> 00000080
// oktal      : 0765      -> fhrendes 'o/O' und Zeichenfolge aus "01234567"
// binr      : 0b010101  -> fhrendes "0b" und Zeichenfolge aus "0..1"
// alles andere wird als ein Textstring erkannt und unverndert bernommen.
//
// Innerhalb von Textstrings wirkt der SEPARATOR (,) natrlich auch als Trennzeichen,
// er teilt den String in zwei Teilstrings und wird dabei selbst unterdrckt.
// Abhilfe:
// Ein ESCAPE (\) direkt vor einem SEPARATOR (,) belt den Separator im Text.
// Ein ESCAPE (\) als erstes Zeichen behandelt den folgenden Teilstring als Zeichenkette.
// Soll das erste Zeichen ein ESCAPE sein, dann mssen 2 ESCAPES gesetzt werden

void parse(char * string, char * result, unsigned int * str_len)
{
int value = 0;
unsigned int i, j,typ, byte_cnt, tmp_char = 0;
unsigned int ready = FALSE, escape = FALSE;
unsigned int p_start, p_end, result_idx, var_len;
char cv[30];

p_start = p_end = result_idx = 0;
result[0] = 0;                         // den String in result[] lschen
var_len = 1;                           // Zahlen in 1 Byte lange Werte umwandeln
byte_cnt = strlen(string);

//printf("\ninput: %s\n", string);

while(1)
   {
// Nach dem ersten Auftreten des Trennzeichnens suchen
// p_end liefert die Position des Trennzeichens
// wenn das Zeichen vor dem Trennzeichen ESCAPE ist,
// dann wird das Trennzeichen berlesen
   for (p_end = p_start; p_end < byte_cnt; p_end++)
      {
      if (string[p_end] == SEPARATOR)
         {
         if (!(escape)) break; // Suche beenden
         else                 // das Escape entfernen
            {
            j = p_end;        // den Rest des Strings nach links schieben
            while( j < (strlen(string) +1))
               {
               string[j - 1] = string[j];
               j++;
               }
            }
         }
      if (string[p_end] == ESCAPE) escape = TRUE; else escape = FALSE;
   }

// wenn das Ende des Strings erreicht ist, dann wurde kein Trennzeichen (mehr) gefunden
// dann kann weiteres Parsen beendet werden
   if (p_end == byte_cnt) ready = TRUE;

// kopiert die zu untersuchende Zeichenkette in den String cv
   for (i = 0; i < (p_end - p_start); i++)
      {
      cv[i] = string[p_start + i];
      }
      cv[i] = 0;           // Terminator anhngen, der Teilstring nun ein Textelement

//printf("Teilstring: %s Laenge: %d\n", cv, strlen(cv));

// Falls der Teilstring leer ist (aufeinanderfolgende Separatoren oder Textanfang/ende)
      if ((strlen(cv) == 0)) // || (strlen(cv) == 1) && (cv[0] == SEPARATOR))
         {

// Falls oben bereits das Textende erkannt wurde
         if (ready)     // ein leerer Teilstring am Ende des Strings
				{
//printf("\nWert bei break: %X - Anzahl Bytes: %X", value, result_idx);
				break;
				}
			else          // ein leerer Teilstring innerhalb des Strings
				{
         	p_start++;
				continue;
				}
			}

// den Startpunkt fr die nchste Suche merken, dazu die (Lnge des Strings + 1)
// (fr das bersprungene Trennzeichen) zu p_start addieren
 		else p_start += (p_end - p_start + 1);

// ein ESCAPE als erste Zeichen verhindert die Interpretation eines String als Zahl
// weiter unten beim Umkopieren wird dann das erste Zeichen nicht kopiert.
		if (cv[0] == ESCAPE) typ = STRING;
		else typ = UNKNOWN;

// printf("Teilstring: %s, Laenge: %d, Typ: %d\n", cv, strlen(cv), typ);

// prfen, ob ein abschlieendes 'S', 'L' oder 'W' angehngt worden ist
// Falls ja, dann werden 8/16/32 Bit Werte zurckgegeben
// Aber da noch nicht bekannt ist, ob berhaupt eine gltige Zahlenfolge vorliegt, wird hier
// das 'l'/'L' abgeschnitten und spter (falls ein Textstring vorliegt) wieder angefgt.
   i =  0;
   while(cv[i]) i++;                      // ans Ende des Textes scrollen
   i--;                                   // zurck an die letzte Textstelle, prfen

//printf("\nTeilstring: %s, Ende: %i", cv, i);

   if ((cv[i] & 0xDF) == FORCE_4_BYTE && (i > 1))
      {
      var_len = 4;                 			// 2 Byte langes Ergebnis
      tmp_char = cv[i];                   // das Zeichen zwischenspeichern
      cv[i] = 0;                          // das 'w/W' durch \0 ersetzen
      }
   else if ((cv[i] & 0xDF) == FORCE_2_BYTE && (i > 1))
      {
      var_len = 2;                 			// Flag setzen
      tmp_char = cv[i];                   // das Zeichen zwischenspeichern
      cv[i] = 0;                          // das 'l/L' durch \0 ersetzen
      }
   else if ((cv[i] & 0xDF) == FORCE_1_BYTE && (i > 1))
      {
      var_len = 1;                 			// Flag setzen
      tmp_char = cv[i];                   // das Zeichen zwischenspeichern
      cv[i] = 0;                          // das 'l/L' durch \0 ersetzen
      }
	else var_len = 0;                      // 0 == noch unbekannte Lnge

//if (tmp_char) printf("\nTmpChar %c %d", tmp_char, i);
//printf("Teilstring: %s Laenge: %d var_len: %d\n", cv, strlen(cv), var_len);

// nun das Format/Zahlentyp des String ermitteln ---------------------------------
// wenn nicht bereits STRING durch ein ESCAPE am Textanfang erzwungen wurde
   if (typ == UNKNOWN) typ = test_numstring(cv);

//  printf("\nTyp: %X\ Idx: %d", typ, result_idx);

// Abhngig vom Ergebnis des Testes werden die Daten nun konvertieren
      if      (typ == UNSIGNED_DEZ) value = dez_2_byte(cv);
      else if (typ == POSITIV_DEZ) value = dez_2_byte(&cv[1]); // das Vorzeichen ignorieren
      else if (typ == NEGATIV_DEZ)          // zuerst den positiven Wert ermitteln
         {                                 // und dann von 0 subtrahieren
         value = 0 - dez_2_byte(&cv[1]);   // das Vorzeichen ignorieren
         }
      else if (typ == HEXADEZIMAL)  value = hex_2_byte(cv);
      else if (typ == BINAER)       value = bin_2_byte(cv);
      else if (typ == OKTAL)        value = oct_2_byte(cv);
//      else if (typ == NOTHING) continue;  // leere Eingabe
      else
          {
//printf("Teilstring: %s, %d\n", cv, strlen(cv));

		// das 1. Zeichen auslassen, wenn es ein ESCAPE ist
			if (cv[0] == ESCAPE) i = 1; else i = 0;

         for (j = i; j < strlen(cv); j++) result[result_idx++] = cv[j];
         if (tmp_char)                    // falls oben 'l/L' abgeschnitten wurde
            {
//printf("\ntemp_char : %c ", tmp_char);
            result[result_idx++] = tmp_char; // 'l/L' wieder anhngen
            tmp_char = 0;
            }
         }

//printf("result_idx: %d\n", result_idx);

// sofern nicht durch Anhngen eines Buchstaben eine feste Datenlnge gefordert wurde
// (wenn nmlich var_len == 0), dann hier die Datenlnge an den Werteumfang anpassen
//
// wenn der Wert nicht mehr in 1 Byte passt, dann automatisch 2 bzw. 4 Byte whlen

//  printf("\nVar_len: %X\ Idx: %d", var_len, result_idx);
      if (var_len == 0)
         {
         if      ((value > 65535) || (value < -32768)) var_len = 4; // Minimum -128 fr 1 Byte
			else if ((value >   255) || (value <   -128)) var_len = 2;
			else var_len = 1;
         }
//  printf("\nVar_len: %X\ Idx: %d", var_len, result_idx);
// Stringfunktionen sind bei Zahleneingaben nicht mglich wegen \0 Bytes innerhalb der Daten !!
      if (!(typ == STRING))
         {
         if (var_len == 4)
            {
            result[result_idx++] = value >> 24;
            result[result_idx++] = value >> 16;
            result[result_idx++] = value >> 8;
				}
			else if (var_len == 2)
				{
            result[result_idx++] = value >> 8;
				}
			result[result_idx++] = value % 256;
         }

      if (ready)
         {
//printf("\nWert ende: %X Idx: %X", value, result_idx);

            break;
         }
   }                                      // while(1)
   *str_len = result_idx;
}

// ----------------------------------------------------------------------------
// testet, in welchem Zahlenformat die Zeichenfolge vorliegt:
// liefert das Ergebnis als Konstante zurck

unsigned int test_numstring(char* s)
{
   char* pt;
   pt = s;                       // Pointer auf Textanfang kopieren
                                 // zuerst auf dezimale Zeichenfolge testen
   while (isdigit(*s)) s++;      // solange wie Ziffern folgen
   if (!(*s)) return (UNSIGNED_DEZ);  // wenn Ende erreicht ist und vorher kein
                                 // Fehler auftrat, dann lag Dezimalzahl vor

   s = pt;                       // zurck zum Anfang der Zeichenfolge

   if (*s == '+')                // ist das 1. Zeichen ein '+' ?
   {                             // wenn ja, dann muss der Rest zwischen 0..7 sein
//   *(++s);
   while (isdigit(*(++s))) s++;      // solange wie Ziffern folgen
   if (!(*s)) return (POSITIV_DEZ); // wenn Ende erreicht ist und vorher kein
                                 // Fehler auftrat, dann lag eine Dezimalzahl mit fhrendem '+' vor
   }

   s = pt;                       // zurck zum Anfang der Zeichenfolge

   if (*s == '-')                // ist das 1. Zeichen ein '-' ?
   {                             // wenn ja, dann muss der Rest zwischen 0..7 sein

 //  *(++s);
   while (isdigit(*(++s))) s++;      // solange wie Ziffern folgen
   if (!(*s)) return (NEGATIV_DEZ);  // wenn Ende erreicht ist und vorher kein
                                 // Fehler auftrat, dann lag eine negative Dezimalzahl vor
   }

   s = pt;                       // zurck zum Anfang der Zeichenfolge

   if (((*s) & 0xDF) == 'O')                // ist das 1. Zeichen ein 'O' ?
   {                             // wenn ja, dann muss der Rest zwischen 0..7 sein
      while (*(++s) >= '0' && *s < '8'); // der Rest "01234567"
      if (!(*s)) return (OKTAL);
   }

   else if (*s == '0')           // wenn das erste Zeichen eine NULL
   {
      if ((*(++s)& 0xDF) == 'X')         // wenn das 2. Zeichen ein X
      {                          // und dann wenn der Rest
         while (*(++s) && isxdigit(*s)); // innerhalb "012...DEF")
         if (!(*s)) return(HEXADEZIMAL);
      }

      else if (((*s) & 0xDF) == 'B')        // wenn das 2. Zeichen ein B
      {                          // und der Rest aus 0 oder 1 besteht
         while (*(++s) >= '0' && *s < '2');
         if (!(*s)) return (BINAER);
      }
   }
   return (STRING);              // dann liegt kein bekanntes Zahlenformat vor
}

// ----------------------------------------------------------------------------
// wandelt eine dezimale Zeichenfolge in eine Zahl um,
// die Gltigkeit der Zeichenfolge wird vorausgesetzt !
unsigned int dez_2_byte(char * cv)
{
unsigned int r = 0;              // Rckgabewert auf NULL setzen

while(*cv)                       // solange bis das Ende des Textstrings erreicht
   {
      r *= 10;                   // das Zwischenergebnis mit 10 multiplizieren
      r += (*cv++ - '0');        // und Einerstellen addieren
   }

return(r);
}


// ----------------------------------------------------------------------------
// wandelt eine binare Zeichenfolge in eine Zahl um,
// die Gltigkeit der Zeichenfolge wird vorausgesetzt !
unsigned int bin_2_byte(char * cv)
{
unsigned int r = 0;              // Rckgabewert auf NULL setzen
cv += 2;                         // die beiden ersten Zeichen '0b' ignorieren
                                 // den Pointer auf s[2] setzen
while(*cv)                       // solange bis das Ende des Textstrings erreicht
   {
      r *= 2;                    // das Ergebnis mit 2 multiplizieren
      if (*cv++ == '1') r++;     // und Einerstelle addieren
   }
return(r);
}

// ----------------------------------------------------------------------------
// wandelt eine hexadezimale Zeichenfolge in eine Zahl um,
// die Gltigkeit der Zeichenfolge wird vorausgesetzt !
unsigned int hex_2_byte(char * cv)
{
unsigned int r = 0;              // Rckgabewert auf NULL setzen

cv += 2;                         // die beiden erste Zeichen '0x' ignorieren
                                 // den Pointer auf s[2] setzen
while(*cv)                       // solange bis das Ende des Textstrings erreicht
   {
   r *= 16;                      // das Zwischenergebnis mit 16 multiplizieren
                                 // und die Einerstellen  addieren

   if (*cv < 'A') r += (*cv++ - '0'); // Ziffern von 0 - 9
   else
      {
      *cv = *cv & 0xDF;          // in Grossschreibung umwandeln
      r += (*cv++ - 'A' + 10);   // Ziffern von A - F (10 - 15)
      }
   }
return(r);
}

// ----------------------------------------------------------------------------
// wandelt eine oktale Zeichenfolge in eine Zahl um,
// die Gltigkeit der Zeichenfolge wird vorausgesetzt !
unsigned int oct_2_byte(char * cv)
{
unsigned int r = 0;              // Rckgabewert auf NULL setzen
cv++;                            // das erste Zeichen 'O' ignorieren
                                 // den Pointer auf s[1] setzen
while(*cv)                       // solange bis das Ende des Textstrings erreicht
   {
      r *= 8;                    // das Ergebnis mit 8 multiplizieren
      r += (*cv++ - '0');        // und Einerstellen addieren
   }
return(r);
}

// --- [eof] ------------------------------------------------------------------
